from .base import Option, Attribute, BaseService, ManagedService, \
    ServiceError, ManagerNotFoundError
from .image import Image, ImageNotFoundError, ImageError, Registry
from .container import Container, ContainerNotFoundError, ContainerError
from .service import Service, ServiceNotFoundError
from .stack import Stack



import ctypes
import functools
import hashlib
import multiprocessing

from fabric import api as fab

import fabricio

from fabricio.utils import DEFAULT


def skip_unknown_host(func):
    @functools.wraps(func)
    def _task(*args, **kwargs):
        if fab.env.get('host_string', False):
            return func(*args, **kwargs)
        fabricio.log(
            "'{func}' execution was skipped due to no host provided "
            "(command: {command})".format(
                func=func.__name__,
                command=fab.env.command,
            )
        )

    _task.wrapped = func  # compatibility with Fabric tasks

    return _task


def once_per_task(func=None, block=False, default=None):
    if func is None:
        return functools.partial(once_per_task, block=block, default=default)

    def wait(timeout=None):
        executed.wait(timeout)

    def set_result(value=DEFAULT):
        if value is not DEFAULT:
            results[get_current_task_id()] = value
        executed.set()

    def reset(block=False):
        lock = last_task.get_lock()
        if lock.acquire(block):
            try:
                if not executed.is_set():
                    return
                current_task = get_current_task_id()
                if current_task != last_task.raw:
                    results.pop(current_task, None)
                    executed.clear()
            finally:
                lock.release()

    def has_result():
        return get_current_task_id() in results

    def get_current_task_id():
        task = fab.env.command or ''
        infrastructure = fab.env.infrastructure or ''
        session = hashlib.md5()
        session.update(task.encode('utf-16be'))
        session.update(infrastructure.encode('utf-16be'))
        for host in fab.env.all_hosts:
            session.update(host.encode('utf-16be'))
        return session.digest()

    @functools.wraps(func)
    def _func(*args, **kwargs):
        lock = last_task.get_lock()
        if lock.acquire(block):
            try:
                current_task = get_current_task_id()
                if current_task != last_task.raw:
                    last_task.raw = current_task
                    results[current_task] = result = func(*args, **kwargs)
                    return result
                return results.get(current_task, default)
            finally:
                executed.set()
                lock.release()

    last_task = multiprocessing.Array(ctypes.c_char, hashlib.md5().digest_size)
    results = multiprocessing.Manager().dict()
    executed = multiprocessing.Event()

    _func.has_result = has_result
    _func.reset = reset
    _func.wait = wait
    _func.set = set_result

    _func.wrapped = func  # compatibility with Fabric tasks

    return _func



import itertools
import multiprocessing
import sys

import six

from cached_property import cached_property
from fabric import api as fab, colors
from frozendict import frozendict

import fabricio

from fabricio import utils

from .image import Image


class ServiceError(fabricio.Error):
    pass


class ManagerNotFoundError(ServiceError):
    pass


class Option(utils.default_property):

    def __init__(
        self,
        func=None,
        default=None,
        name=None,
        safe=False,
        safe_name=None,
    ):
        super(Option, self).__init__(func=func, default=default)
        self.name = name
        self.safe = safe
        self.safe_name = safe_name


class Attribute(utils.default_property):
    pass


class BaseService(object):

    image = Image()

    @Attribute
    def name(self):
        raise ValueError('must provide service name')

    def __init__(self, image=None, options=None, safe_options=None, **attrs):
        if image is not None:
            self.image = image
        self.overridden_options = set()
        self.overridden_attributes = set()
        is_attribute = self._attributes.__contains__
        self._other_options = other_options = options or {}
        self._other_safe_options = safe_options or {}
        for option in list(other_options):
            attr = self._get_option_attribute(option)
            if attr:
                setattr(self, attr, other_options.pop(option))
        for attr, value in attrs.items():
            if not is_attribute(attr):
                raise TypeError('Unknown attribute: {attr}'.format(attr=attr))
            setattr(self, attr, value)

    def __setattr__(self, attr, value):
        if attr in self._options:
            self.overridden_options.add(attr)
        elif attr in self._attributes:
            self.overridden_attributes.add(attr)
        super(BaseService, self).__setattr__(attr, value)

    def _get_option_attribute(self, option_name):
        if option_name in self._options:
            return option_name
        if option_name in self._options_custom_names:
            return self._options_custom_names[option_name]
        return None

    @cached_property
    def _options_custom_names(self):
        return dict(
            (option.name, attr)
            for attr, option in self._options.items()
            if option.name
        )

    @property
    def image_id(self):
        raise NotImplementedError

    @cached_property
    def _attributes(self):
        return set(
            attr
            for cls in type(self).__mro__[::-1]
            for attr, value in vars(cls).items()
            if isinstance(value, Attribute)
        )

    @classmethod
    def _get_available_options(cls, safe=False):
        return dict(
            (attr, option)
            for mro in cls.__mro__[::-1]
            for attr, option in vars(mro).items()
            if isinstance(option, Option)
            and (safe and option.safe or option.safe_name or not safe)
        )

    @cached_property
    def _options(self):
        return self._get_available_options()

    @cached_property
    def _safe_options(self):
        return self._get_available_options(safe=True)

    def _get_options(self, safe=False):
        options = itertools.chain(
            (
                (
                    safe and option.safe_name or option.name or attr,
                    getattr(self, attr),
                )
                for attr, option in
                (self._safe_options if safe else self._options).items()
            ),
            (self._other_safe_options if safe else self._other_options).items(),
        )
        evaluated_options = (
            (option, value(self) if callable(value) else value)
            for option, value in options
        )
        return frozendict(
            (option, value)
            for option, value in evaluated_options
            if value is not None
        )

    @property
    def options(self):
        return self._get_options()

    @property
    def safe_options(self):
        return self._get_options(safe=True)

    def fork(self, image=None, options=None, **attrs):
        image = image or self.image
        fork_options = dict(
            (
                (option, getattr(self, option))
                for option in self.overridden_options
            ),
            **self._other_options
        )
        if options:
            fork_options.update(options)
        if self.overridden_attributes:
            attrs = dict(
                (
                    (attr, getattr(self, attr))
                    for attr in self.overridden_attributes
                ),
                **attrs
            )
        return self.__class__(image=image, options=fork_options, **attrs)

    def __str__(self):
        return six.text_type(self.name)

    def __copy__(self):
        return self.fork()

    def get_backup_version(self):
        raise NotImplementedError

    def update(self, tag=None, registry=None, account=None, force=False):
        raise NotImplementedError

    def revert(self):
        raise NotImplementedError

    def pull_image(self, tag=None, registry=None, account=None):
        image = self.image[registry:tag:account]
        if image:
            return image.pull()

    def migrate(self, tag=None, registry=None, account=None):
        pass

    def migrate_back(self):
        pass

    def backup(self):
        pass

    def restore(self, backup_name=None):
        pass

    def destroy(self):
        raise NotImplementedError

    @property
    def info(self):
        raise NotImplementedError


class ManagedService(BaseService):

    def __init__(self, *args, **kwargs):
        super(ManagedService, self).__init__(*args, **kwargs)
        self.managers = multiprocessing.Manager().dict()

    def _is_manager(self):
        command = 'docker info 2>&1 | grep "Is Manager:"'
        return fabricio.run(command).endswith('true')

    def is_manager(self, raise_manager_error=True):
        is_manager = self.managers.get(fab.env.host)
        try:
            if is_manager is None:
                is_manager = self.managers[fab.env.host] = self._is_manager()
        except fabricio.host_errors as error:
            is_manager = self.managers[fab.env.host] = False
            fabricio.log(
                'WARNING: {error}'.format(error=error),
                output=sys.stderr,
                color=colors.red,
            )
        finally:
            if (
                raise_manager_error
                and len(self.managers) >= len(fab.env.all_hosts)
                and not any(self.managers.values())
            ):
                msg = 'service manager not found or it failed to pull image'
                raise ManagerNotFoundError(msg)
        return is_manager

    def pull_image(self, *args, **kwargs):
        try:
            return super(ManagedService, self).pull_image(*args, **kwargs)
        except fabricio.host_errors as error:
            self.managers[fab.env.host] = False
            fabricio.log(
                'WARNING: {error}'.format(error=error),
                output=sys.stderr,
                color=colors.red,
            )

    def migrate(self, *args, **kwargs):
        if self.is_manager():
            super(ManagedService, self).migrate(*args, **kwargs)

    def migrate_back(self):
        if self.is_manager():
            super(ManagedService, self).migrate_back()

    def backup(self):
        if self.is_manager():
            super(ManagedService, self).backup()

    def restore(self, backup_name=None):
        if self.is_manager():
            super(ManagedService, self).restore(backup_name=backup_name)



import functools
import hashlib
import itertools
import json
import re
import shlex

import dpath
import six

from cached_property import cached_property
from frozendict import frozendict
from six.moves import map, shlex_quote, range

import fabricio

from fabricio import utils

from .base import ManagedService, Option, Attribute, ServiceError


def get_option_value(string, option):
    """
    get option value from string with nested options
    such as "source=from,destination=to"
    """
    for part in re.split(r'(?<!\\),', string):
        if not part.startswith(option + '='):
            continue
        value = shlex.split(part.split('=', 1)[-1])
        return value and value[0] or None


class ServiceNotFoundError(ServiceError):
    pass


class RemovableOption(Option):

    path = NotImplemented

    force_add = False

    force_rm = False

    def cast_rm(self, value):
        return value

    def __init__(self, func=None, path=None, force_add=None, force_rm=None, **kwargs):  # noqa
        super(RemovableOption, self).__init__(func=func, **kwargs)
        self.path = path or self.path
        self.force_add = force_add if force_add is not None else self.force_add
        self.force_rm = force_rm if force_rm is not None else self.force_rm

    def get_values_to_add(self, service, attr):
        new_values = self.get_new_values(service, attr)
        new_values = utils.OrderedSet(new_values)

        if not self.force_add:
            current_values = self.get_current_values(service.info)
            new_values -= current_values

        return list(new_values) or None

    def get_values_to_remove(self, service, attr):
        current_values = self.get_current_values(service.info)
        current_values = utils.OrderedSet(current_values)

        if not self.force_rm:
            new_values = self.get_new_values(service, attr)
            new_values = map(self.cast_rm, new_values)
            new_values = (
                value
                for values in new_values
                for value in (
                    values if isinstance(values, (list, tuple)) else [values]
                )
            )
            current_values -= new_values

        return list(current_values) or None

    def get_current_values(self, service_info):
        if '*' in self.path:
            return dpath.util.values(service_info, self.path)
        try:
            return dpath.util.get(service_info, self.path)
        except KeyError:
            return []

    @staticmethod
    def get_new_values(service, attr):
        values = getattr(service, attr)
        if callable(values):
            values = values(service)
        if values is None:
            return []
        if isinstance(values, six.string_types):
            return [values]
        return values


class LabelOption(RemovableOption):

    def cast_rm(self, value):
        # 'ENV=value' => 'ENV'
        return value.split('=', 1)[0]


class EnvOption(LabelOption):

    path = '/Spec/TaskTemplate/ContainerSpec/Env/*'

    def get_current_values(self, service_info):
        values = super(EnvOption, self).get_current_values(service_info)
        return map(lambda value: value.split('=', 1)[0], values)


class PublishOption(RemovableOption):

    path = '/Spec/EndpointSpec/Ports/*/TargetPort'

    force_add = True

    def cast_rm(self, value, from_current=False):
        # 'source:target/protocol' => int('target')
        target_port = six.text_type(value).rsplit('/', 1)[0].rsplit(':', 1)[-1]
        if '-' not in target_port:
            return int(target_port)
        port_start, port_end = map(int, target_port.split('-', 1))
        return list(range(port_start, port_end + 1))


class MountOption(RemovableOption):

    path = '/Spec/TaskTemplate/ContainerSpec/Mounts/*/Target'

    force_add = True

    def cast_rm(self, value):
        # 'type=volume,destination=/path' => '/path'
        destination = get_option_value(value, 'destination')
        destination = destination or get_option_value(value, 'target')
        return destination or get_option_value(value, 'dst')


class HostOption(RemovableOption):

    path = '/Spec/TaskTemplate/ContainerSpec/Hosts/*'

    def get_current_values(self, *args, **kwargs):
        values = super(HostOption, self).get_current_values(*args, **kwargs)
        # 'ip host' => 'host:ip'
        return map(lambda value: ':'.join(value.split(' ', 1)[::-1]), values)


class PlacementPrefOption(RemovableOption):

    path = '/Spec/TaskTemplate/Placement/Preferences/*'

    force_add = True

    force_rm = True

    def get_current_values(self, *args, **kwargs):
        values = super(PlacementPrefOption, self).get_current_values(*args, **kwargs)  # noqa
        for value in values:
            yield ','.join(
                '{strategy}={descriptor}'.format(
                    strategy=strategy.lower(),
                    descriptor=shlex_quote(descriptor[strategy + 'Descriptor'])
                )
                for strategy, descriptor in value.items()
            )


class Service(ManagedService):

    options_label_name = 'fabricio.service.options'

    command = Attribute()
    args = Attribute()
    mode = Attribute()

    env = EnvOption(safe=True)
    label = LabelOption(path='/Spec/Labels')
    container_label = LabelOption(
        name='container-label',
        path='/Spec/TaskTemplate/ContainerSpec/Labels',
        safe_name='label',
    )
    publish = PublishOption()
    constraint = RemovableOption(
        path='/Spec/TaskTemplate/Placement/Constraints/*',
    )
    replicas = Option()
    mount = MountOption(safe=True)
    network = RemovableOption(
        path='/Spec/TaskTemplate/Networks/*/Target',
        force_add=True,
        force_rm=True,
        safe=True,
    )
    restart_condition = Option(name='restart-condition')
    stop_grace_period = Option(name='stop-grace-period')
    user = Option(safe=True)
    host = HostOption(safe_name='add-host')
    secret = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/Secrets/*/SecretName',
        force_add=True,
        force_rm=True,
    )
    config = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/Configs/*/ConfigName',
        force_add=True,
        force_rm=True,
    )
    group = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/Groups/*',
        safe_name='group-add',
    )
    placement_pref = PlacementPrefOption(name='placement-pref')
    dns = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/DNSConfig/Nameservers/*',
        safe=True,
    )
    dns_option = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/DNSConfig/Options/*',
        name='dns-option',
        safe=True,
    )
    dns_search = RemovableOption(
        path='/Spec/TaskTemplate/ContainerSpec/DNSConfig/Search/*',
        name='dns-search',
        safe=True,
    )

    @property
    def cmd(self):
        return ' '.join([self.command or '', self.args or '']).strip()

    @utils.default_property
    def image_id(self):
        return self.info['Spec']['TaskTemplate']['ContainerSpec']['Image']

    def get_backup_version(self):
        current_info = self.info
        if 'PreviousSpec' not in current_info:
            raise ServiceError('service backup not found')
        previous_spec = current_info['PreviousSpec']
        backup_version = self.fork()
        backup_version.info = dict(self.info, Spec=previous_spec)
        return backup_version

    @cached_property
    def _update_options(self):
        options = {}
        for attr, option in self._options.items():
            name = option.name or attr
            if isinstance(option, RemovableOption):
                options[name + '-rm'] = functools.partial(
                    option.get_values_to_remove,
                    attr=attr,
                )
                options[name + '-add'] = functools.partial(
                    option.get_values_to_add,
                    attr=attr,
                )
            else:
                options[name] = getattr(self, attr)
        return options

    @property
    def update_options(self):
        options = itertools.chain(
            (
                (option, value)
                for option, value in self._update_options.items()
            ),
            self._other_options.items(),
        )
        evaluated_options = (
            (option, value(self) if callable(value) else value)
            for option, value in options
        )
        return frozendict(
            (
                (option, value)
                for option, value in evaluated_options
                if value is not None
            ),
            args=self.cmd,
        )

    def _update_service(self, options):
        fabricio.run('docker service update {options} {service}'.format(
            options=options,
            service=self,
        ))

    def _create_service(self, image):
        command = 'docker service create {options} {image} {cmd}'
        fabricio.run(command.format(
            options=utils.Options(self.options, name=self, mode=self.mode),
            image=image,
            cmd=self.cmd,
        ))

    @fabricio.once_per_task
    def _update(self, image, force=False):
        image = image.digest
        try:
            service_info = self.info
        except ServiceNotFoundError:
            service_info = {}

        with utils.patch(self, 'info', service_info, force_delete=True):
            labels = service_info.get('Spec', {}).get('Labels', {})
            current_options = labels.pop(self.options_label_name, None)
            new_options = self._encode_options(dict(
                self.options,
                image=image,
                args=self.cmd,
            ))

            if force or current_options != new_options:
                label_with_new_options = {
                    self.options_label_name: new_options,
                }
                self._update_labels(label_with_new_options)

                if service_info:
                    options = utils.Options(self.update_options, image=image)
                    self._update_service(options)
                else:
                    self._create_service(image)

                return True
        return False

    def update(self, tag=None, registry=None, account=None, force=False):
        if not self.is_manager():
            return False
        result = self._update(self.image[registry:tag:account], force=force)
        return result is None or result

    @fabricio.once_per_task
    def _revert(self):
        command = 'docker service rollback {service}'.format(service=self)
        fabricio.run(command)

    def revert(self):
        if self.is_manager():
            self._revert()

    @utils.default_property
    def info(self):
        command = 'docker service inspect {service}'
        info = fabricio.run(
            command.format(service=self),
            abort_exception=ServiceNotFoundError,
        )
        return json.loads(info)[0]

    @staticmethod
    def _encode_options(options):
        bucket = json.dumps(options, sort_keys=True, default=six.text_type)
        return hashlib.md5(bucket.encode()).hexdigest()

    def _update_labels(self, labels):
        service_labels = self.label
        if not service_labels:
            service_labels = []
        elif isinstance(service_labels, six.string_types):
            service_labels = [service_labels]
        else:
            try:
                service_labels = list(service_labels)
            except TypeError:
                service_labels = [six.text_type(service_labels)]
        for label, value in labels.items():
            service_labels.append("{0}={1}".format(label, value))
        self.label = service_labels

    def destroy(self, **options):
        """
        any passed argument will be forwarded to 'docker service rm' as option
        """
        if self.is_manager():
            self._destroy(**options)

    @fabricio.once_per_task
    def _destroy(self, **options):
        fabricio.run('docker service rm {options} {name}'.format(
            options=utils.Options(options),
            name=self.name,
        ))



from six.moves import filter, reduce, map

import fabricio

from fabricio import docker, utils


class Configuration(docker.Stack):

    get_update_command = 'kubectl apply {options}'.format

    @property
    def current_settings_tag(self):
        return 'fabricio-current-kubernetes:{0}'.format(self.name)

    @property
    def backup_settings_tag(self):
        return 'fabricio-backup-kubernetes:{0}'.format(self.name)

    @docker.Option(name='filename')
    def config(self):
        raise docker.ServiceError("must provide 'config' or 'filename' option")

    def _is_manager(self):
        return fabricio.run(
            'kubectl config current-context',
            ignore_errors=True,
        ).succeeded

    def _revert_images(self, digests):
        spec = self.__get_images_spec()
        for kind, images in spec.items():
            image_updates = ' '.join(
                '{0}={1}'.format(name, digests[image])
                for name, image in images.items()
            )
            command = 'kubectl set image {kind} {images}'
            command = command.format(kind=kind, images=image_updates)
            fabricio.run(command)

    @property
    @fabricio.once_per_task(block=True)
    def images(self):
        with self.upload_configuration_file():
            spec = self.__get_images_spec()
        return list(reduce(set.union, map(dict.values, spec.values()), set()))

    def __get_images_spec(self):
        template = (  # noqa
            '{{define "images"}}'
                '{{$kind := .kind}}'
                '{{$name := .metadata.name}}'
                '{{with .spec.template.spec.containers}}'
                    '{{range .}}'
                        r'{{$kind}}/{{$name}} {{.name}} {{.image}}{{"\n"}}'
                    '{{end}}'
                '{{end}}'
            '{{end}}'
            '{{if eq .kind "List"}}'
                '{{range .items}}{{template "images" .}}{{end}}'
            '{{else}}'
                '{{template "images" .}}'
            '{{end}}'
        )
        command = (
            'kubectl get {options}'.format(options=utils.Options([
                ('output', 'go-template'),
                ('filename', self.config),
                ('template', template),
            ]))
        )

        result = dict()
        for line in filter(None, fabricio.run(command).splitlines()):
            kind, image_spec = line.split(None, 1)
            name, image = image_spec.rsplit(None, 1)
            result.setdefault(kind, dict())[name] = image

        return result

    def destroy(self, **options):
        """
        any passed argument will be forwarded to 'kubectl delete' as option

        Note: make sure "managers" are listed before "workers" in your
        Fabricio configuration before calling this method in serial mode
        """
        with self.upload_configuration_file():
            super(Configuration, self).destroy(**options)

    @fabricio.once_per_task(block=True)
    def _destroy(
        self,
        options,  # type: utils.Options
    ):
        self.images  # get list of images before configuration remove
        options.setdefault('filename', self.config)
        fabricio.run('kubectl delete {options}'.format(options=options))



import collections
import contextlib
import warnings

from distutils import util as distutils

import six

from six.moves import shlex_quote

import fabricio

DEFAULT = object()


@contextlib.contextmanager
def patch(obj, attr, value, default=DEFAULT, force_delete=False):
    original = not force_delete and getattr(obj, attr, default)
    setattr(obj, attr, value)
    try:
        yield
    finally:
        if force_delete or original is DEFAULT:
            obj.__delattr__(attr)
        else:
            setattr(obj, attr, original)


class default_property(object):

    def __init__(self, func=None, default=None):
        self.func = func
        self.default = default

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.func is None:
            return self.default
        return self.func(instance)

    def __call__(self, func):
        self.func = func
        return self


class Options(collections.OrderedDict):

    def make_option(self, option, value=None):
        option = '--' + option
        if value is not None:
            option += '=' + shlex_quote(six.text_type(value))
        return option

    def make_options(self):
        for option, value in self.items():
            if value is None:
                continue
            if isinstance(value, bool):
                if value is True:
                    yield self.make_option(option)
            elif isinstance(value, six.string_types):
                yield self.make_option(option, value)
            elif isinstance(value, collections.Iterable):
                for single_value in value:
                    yield self.make_option(option, single_value)
            else:
                yield self.make_option(option, value)

    def __str__(self):
        return ' '.join(self.make_options())


def strtobool(value):
    return distutils.strtobool(six.text_type(value))


def once_per_command(*args, **kwargs):  # pragma: no cover
    warnings.warn(
        'once_per_command renamed to fabricio.decorators.once_per_task, '
        'this one will be removed in 0.6',
        DeprecationWarning,
    )
    warnings.warn(
        'once_per_command renamed to fabricio.decorators.once_per_task, '
        'this one will be removed in 0.6',
        RuntimeWarning, stacklevel=2,
    )
    return fabricio.once_per_task(*args, **kwargs)


class AttrDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


class PriorityDict(dict):

    def __init__(self, iterable=None, priority=None, **kwargs):
        super(PriorityDict, self).__init__(iterable, **kwargs)
        self.priority = priority or []

    def items(self):
        seen = set()
        for key in self.priority:
            if key in self:
                yield key, self[key]
                seen.add(key)
        for key in self:
            if key not in seen:
                yield key, self[key]


class OrderedSet(collections.MutableSet):  # pragma: no cover
    """
    http://code.activestate.com/recipes/576694-orderedset/
    """

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)

    # additional methods

    union = collections.MutableSet.__ior__



import itertools
import json
import sys
import warnings

from base64 import b64encode, b64decode

import contextlib2 as contextlib
import six

from fabric import colors, api as fab
from six.moves import map, shlex_quote, filter, zip_longest

import fabricio

from fabricio import utils

from .base import ManagedService, Option, Attribute, ServiceError, \
    ManagerNotFoundError
from .image import Image, ImageNotFoundError


class Stack(ManagedService):

    image_id = None

    info = None

    temp_dir = Attribute(default='/tmp')

    config = Option(name='compose-file', default='docker-compose.yml')

    @property
    def compose_file(self):  # pragma: no cover
        warnings.warn(
            "'compose_file' option is deprecated and will be removed in v0.6, "
            "use 'config' or 'compose-file' instead", DeprecationWarning,
        )
        return self.config

    configuration_label = 'fabricio.configuration'

    digests_label = 'fabricio.digests'

    get_update_command = 'docker stack deploy {options} {name}'.format

    def __init__(self, *args, **kwargs):
        options = kwargs.setdefault('options', {})
        if 'compose_file' in options:  # pragma: no cover
            warnings.warn(
                "'compose_file' option is deprecated and will be removed"
                " in v0.6, use 'config' or 'compose-file' instead",
                RuntimeWarning, stacklevel=2,
            )
            options.setdefault('config', options['compose_file'])

        super(Stack, self).__init__(*args, **kwargs)
        self._current_configuration = None

    @property
    def current_settings_tag(self):
        return 'fabricio-current-stack:{0}'.format(self.name)

    @property
    def backup_settings_tag(self):
        return 'fabricio-backup-stack:{0}'.format(self.name)

    @contextlib.contextmanager
    def upload_configuration_file(self, configuration=None):
        if self._current_configuration is not None or not self.is_manager():
            yield self._current_configuration
        else:
            with fab.cd(self.temp_dir):
                try:
                    configuration = configuration or self.get_configuration()
                    self._current_configuration = configuration
                    fab.put(six.BytesIO(configuration), self.config)
                    yield configuration
                finally:
                    fabricio.remove_file(self.config, ignore_errors=True)
                    self._current_configuration = None

    def upload_configuration(self, configuration):  # pragma: no cover
        warnings.warn(
            'this method is deprecated and will be removed in v0.6, '
            'use upload_configuration_file context manager instead',
            DeprecationWarning,
        )
        warnings.warn(
            'upload_configuration is deprecated and will be removed in v0.6, '
            'use upload_configuration_file context manager instead',
            RuntimeWarning, stacklevel=2,
        )
        fab.put(six.BytesIO(configuration), self.config)

    def get_configuration(self):
        return open(self.config, 'rb').read()

    def update(self, tag=None, registry=None, account=None, force=False):
        if not self.is_manager():
            return None

        with self.upload_configuration_file() as configuration:
            updated = self._update(configuration, force=force)

            if updated:
                self.save_new_settings(
                    configuration=configuration,
                    image=self.image[registry:tag:account],
                )

        return updated

    @fabricio.once_per_task(block=True)
    def _update(self, new_configuration, force=False):
        if not force:
            configuration, digests = self.current_settings
            if configuration == new_configuration and digests is not None:
                new_digests = self._get_digests(digests)
                if digests == new_digests:
                    return False

        options = utils.Options(self.options)
        command = self.get_update_command(options=options, name=self.name)
        fabricio.run(command)

        return True

    def revert(self):
        if not self.is_manager():
            return
        self._revert()
        if self._revert.has_result():
            self.rotate_sentinel_images(rollback=True)

    @fabricio.once_per_task(block=True)
    def _revert(self):
        configuration, digests = self.backup_settings

        if configuration is None:
            raise ServiceError('backup configuration not found')

        with self.upload_configuration_file(configuration):
            self._update(configuration, force=True)

            if digests:
                self._revert_images(digests)

    def _revert_images(self, digests):
        images = self.__get_images()
        for service, image in images.items():
            digest = digests[image]
            command = 'docker service update --image {digest} {service}'
            command = command.format(digest=digest, service=service)
            fabricio.run(command)

    @property
    def current_settings(self):
        return self._get_settings(Image(self.current_settings_tag))

    @property
    def backup_settings(self):
        return self._get_settings(Image(self.backup_settings_tag))

    def _get_settings(self, image):
        try:
            labels = image.info.get('Config', {}).get('Labels', {})
            configuration = labels.get(self.configuration_label)
            configuration = configuration and b64decode(configuration)
            digests = labels.get(self.digests_label)
            digests = digests and json.loads(b64decode(digests).decode())
            return configuration, digests
        except ImageNotFoundError:
            return None, None

    def rotate_sentinel_images(self, rollback=False):
        backup_tag = self.backup_settings_tag
        current_tag = self.current_settings_tag
        if rollback:
            backup_tag, current_tag = current_tag, backup_tag

        backup_images = [backup_tag]
        with contextlib.suppress(ImageNotFoundError):
            backup_images.append(Image(backup_tag).info['Parent'])

        with contextlib.suppress(fabricio.host_errors):
            # TODO make separate call for each docker command
            fabricio.run(
                (
                    'docker rmi {backup_images}'
                    '; docker tag {current_tag} {backup_tag}'
                    '; docker rmi {current_tag}'
                ).format(
                    backup_images=' '.join(backup_images),
                    current_tag=current_tag,
                    backup_tag=backup_tag,
                ),
            )

    def save_new_settings(self, configuration, image):
        self.rotate_sentinel_images()

        labels = [(self.configuration_label, b64encode(configuration).decode())]
        with contextlib.suppress(fabricio.host_errors):
            digests = self._get_digests(self.images)
            digests_bucket = json.dumps(digests, sort_keys=True)
            digests_bucket = b64encode(digests_bucket.encode()).decode()
            labels.append((self.digests_label, digests_bucket))

        dockerfile = (
            'FROM {image}\n'
            'LABEL {labels}\n'
        ).format(
            image=image or 'scratch',
            labels=' '.join(itertools.starmap('{0}={1}'.format, labels)),
        )
        build_command = 'echo {dockerfile} | docker build --tag {tag} -'.format(
            dockerfile=shlex_quote(dockerfile),
            tag=self.current_settings_tag,
        )

        try:
            fabricio.run(build_command)
        except fabricio.host_errors as error:
            fabricio.log(
                'WARNING: {error}'.format(error=error),
                output=sys.stderr,
                color=colors.red,
            )

    @property
    @fabricio.once_per_task(block=True)
    def images(self):
        images = self.__get_images()
        return list(set(images.values()))

    def __get_images(self):
        command = 'docker stack services --format "{{.Name}} {{.Image}}" %s'
        command %= self.name
        lines = filter(None, fabricio.run(command).splitlines())
        return dict(map(lambda line: line.rsplit(None, 1), lines))

    @staticmethod
    def _get_digests(images):
        if not images:
            return {}

        for image in images:
            Image(image).pull(use_cache=True, ignore_errors=True)

        command = (
            'docker inspect --type image --format "{{index .RepoDigests 0}}" %s'
        ) % ' '.join(images)
        digests = fabricio.run(command, ignore_errors=True, use_cache=True)

        return dict(zip_longest(images, filter(None, digests.splitlines())))

    def get_backup_version(self):
        return self.fork(image=self.backup_settings_tag)

    def destroy(self, **options):
        """
        any passed argument will be forwarded to 'docker stack rm' as option

        Note: make sure "managers" are listed before "workers" in your
        Fabricio configuration before calling this method in serial mode
        """
        self._destroy.reset(block=True)

        try:
            if self.is_manager():
                self._destroy(utils.Options(options))
        except ManagerNotFoundError:
            self._destroy.set()
            raise

        timeout = None if fab.env.parallel else 0
        self._destroy.wait(timeout)

        if self._destroy.has_result():
            self._remove_images()

    @fabricio.once_per_task(block=True)
    def _destroy(
        self,
        options,  # type: utils.Options
    ):
        self.images  # get list of images before stack remove
        fabricio.run('docker stack rm {options} {name}'.format(
            options=options,
            name=self.name,
        ))

    def _remove_images(self):
        images = [self.current_settings_tag, self.backup_settings_tag]
        with contextlib.suppress(ImageNotFoundError):
            images.append(Image(self.current_settings_tag).info['Parent'])
            images.append(Image(self.backup_settings_tag).info['Parent'])
        images.extend(self.images)
        fabricio.run(
            'docker rmi {images}'.format(images=' '.join(images)),
            ignore_errors=True,
        )



import functools
import os
import sys
import types
import warnings

import contextlib2 as contextlib
import six

from fabric import api as fab, colors, state
from fabric.contrib import console
from fabric.main import is_task_object
from fabric.tasks import Task, WrappedCallableTask, get_task_details

import fabricio

from fabricio import docker, utils
from fabricio.misc import dangling_images_delete_command

fab.env.setdefault('infrastructure', None)


# add to Fabric's WrappedCallableTask additional method to get
# benefits from using super() inside method-based tasks
WrappedCallableTask.__get__ = lambda self, instance, owner: (
    self.__class__(
        self.wrapped.__get__(instance, owner),
        aliases=self.aliases,
        default=self.is_default,
        name=self.name,
    )
)


def get_task_name(task, _cache={}):
    if not _cache:
        def _fill_cache(mapping=state.commands, keys=[]):
            for key, value in mapping.items():
                _keys = keys + [key]
                if isinstance(value, dict):
                    _fill_cache(value, _keys)
                else:
                    _cache[value] = '.'.join(_keys)
        _fill_cache()
    return _cache.get(task)


def is_task_mode():
    # if there is at least one task to run then assume it is task mode
    return bool(fab.env.tasks)


def execute(*args, **kwargs):  # pragma: no cover
    warnings.warn(
        'fabricio.execute() is deprecated in favour of fabric.api.execute()',
        DeprecationWarning,
    )
    warnings.warn(
        'fabricio.execute() is deprecated and will be removed in v0.6, '
        'use fabric.api.execute() instead',
        RuntimeWarning, stacklevel=2,
    )
    try:
        task, args = args[0], args[1:]
    except IndexError:
        raise TypeError('must provide task to execute')
    default_name = '{command}.{task_name}({id})'.format(
        command=fab.env.command,
        task_name=getattr(task, 'name', task.__name__),
        id=id(task),
    )
    with utils.patch(task, 'name', get_task_name(task) or default_name):
        return fab.execute(task, *args, **kwargs)


def skip_unknown_host(*args, **kwargs):  # pragma: no cover
    warnings.warn(
        'skip_unknown_host decorator moved to fabricio.decorators module,'
        'this one will be removed in 0.6',
        DeprecationWarning,
    )
    warnings.warn(
        'skip_unknown_host decorator moved to fabricio.decorators module'
        'this one will be removed in 0.6',
        RuntimeWarning, stacklevel=2,
    )
    return fabricio.skip_unknown_host(*args, **kwargs)


class SshTunnel(object):

    bind_address = '127.0.0.1'

    host = 'localhost'

    def __new__(cls, mapping):
        if mapping is None:
            return None
        return super(SshTunnel, cls).__new__(cls)

    def __init__(self, mapping):
        mapping = six.text_type(mapping)
        parts = mapping.split(':', 3)
        if len(parts) == 4:
            self.bind_address, port, self.host, host_port = parts
        elif len(parts) == 3:
            port, self.host, host_port = parts
        elif len(parts) == 2:
            port, host_port = parts
        else:
            port = host_port = parts[0]
        self.port = int(port)
        self.host_port = int(host_port)


class Tasks(object):

    @utils.default_property
    def __name__(self):
        return self

    __class__ = types.ModuleType

    def __new__(cls, *args, **kwargs):
        self = super(Tasks, cls).__new__(cls)
        for attr in dir(cls):
            obj = getattr(self, attr, None)
            if is_task_object(obj):
                setattr(self, attr, obj)
        return self

    def __init__(self, roles=(), hosts=(), create_default_roles=True):
        if create_default_roles:
            for role in roles:
                fab.env.roledefs.setdefault(role, [])
        for task in self:
            if not hasattr(task, 'roles'):
                task.roles = roles
            if not hasattr(task, 'hosts'):
                task.hosts = hosts

    def __iter__(self):
        for obj in vars(self).values():
            if is_task_object(obj):
                yield obj


class Infrastructure(Task, Tasks):
    """
    select {title} infrastructure, '{name}.confirm' skips confirmation dialog
    """

    def __new__(cls, callback=None, **kwargs):
        if not callback:
            return functools.partial(cls, **kwargs)
        return super(Infrastructure, cls).__new__(cls)

    def __init__(self, callback, color=colors.yellow, title=None):
        super(Infrastructure, self).__init__()

        # We need to be sure that `default()` will be at first place
        # every time when vars(self) is being invoked.
        # This is necessary to exclude `default` from the list of task
        # because of it's already there as default task.
        # See `fabric.main.extract_tasks()` for details
        self.__dict__ = utils.PriorityDict(self.__dict__, priority=['default'])

        default, confirm = self.default, self.confirm

        self.callback = callback
        self.name = getattr(callback, '__name__', self.name)
        self.title = color(title or self.name)
        self.__doc__ = (self.__doc__ or '').format(**self.__dict__)
        default.__doc__ = (default.__doc__ or '').format(**self.__dict__)
        confirm.__doc__ = (confirm.__doc__ or '').format(**self.__dict__)

        # enable "module" behaviour in task mode
        self.use_task_objects = not is_task_mode()

    @fab.task(
        default=True,
        # mock another task name to exclude this task from the tasks list
        name='confirm',
    )
    @fab.serial
    def default(self, *args, **kwargs):
        """
        select {title} infrastructure
        """
        with utils.patch(fab.env, 'parallel', False):
            if not console.confirm(
                'Are you sure you want to select {title} '
                'infrastructure to run task(s) on?'.format(title=self.title),
                default=False,
            ):
                fab.abort('Aborted')
        return self.confirm(*args, **kwargs)

    @fab.task
    @fab.serial
    def confirm(self, *args, **kwargs):
        """
        select {title} infrastructure skipping confirmation dialog
        """
        fab.env.infrastructure = self.name
        return self.run(*args, **kwargs)

    def run(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def __details__(self):
        doc = self.__doc__ + (self.callback.__doc__ or '')
        with utils.patch(self.callback, '__doc__', doc):
            return get_task_details(self.callback)


infrastructure = Infrastructure


class DockerTasks(Tasks):

    _warnings_stacklevel = 2

    def __init__(
        self,
        service=None,
        registry=None,
        host_registry=None,
        account=None,
        ssh_tunnel_port=None,  # deprecated
        ssh_tunnel=None,
        migrate_commands=False,
        backup_commands=False,
        pull_command=False,
        update_command=False,
        revert_command=False,
        destroy_command=False,
        **kwargs
    ):
        self.destroy = self.DestroyTask(tasks=self)

        super(DockerTasks, self).__init__(**kwargs)

        # We need to be sure that `deploy()` will be at first place
        # every time when vars(self) is being invoked.
        # This is necessary to exclude `deploy` from the list of task
        # because of it's already there as default task.
        # See `fabric.main.extract_tasks()` for details
        self.__dict__ = utils.PriorityDict(self.__dict__, priority=['deploy'])

        # reassign is_default property which is being lost
        # after applying @fab.hosts/@fab.roles decorators
        self.deploy.is_default = True

        self.service = service
        self.registry = registry
        self.host_registry = host_registry or registry
        self.account = account
        self.ssh_tunnel = ssh_tunnel

        if ssh_tunnel_port:
            warnings.warn(
                'ssh_tunnel_port is deprecated and will be removed in v0.6, '
                'use ssh_tunnel instead',
                RuntimeWarning, stacklevel=self._warnings_stacklevel,
            )
            registry = self.registry or self.image and self.image.registry
            assert registry, 'must provide registry if using ssh_tunnel_port'
            self.host_registry = 'localhost:%d' % ssh_tunnel_port
            self.ssh_tunnel = '{port}:{host}:{host_port}'.format(
                port=ssh_tunnel_port,
                host=registry.host,
                host_port=registry.port,
            )

        task_mode = is_task_mode()
        self.backup.use_task_objects = task_mode or backup_commands
        self.restore.use_task_objects = task_mode or backup_commands
        self.migrate.use_task_objects = task_mode or migrate_commands
        self.migrate_back.use_task_objects = task_mode or migrate_commands
        self.revert.use_task_objects = task_mode or revert_command
        self.pull.use_task_objects = task_mode or pull_command
        self.update.use_task_objects = task_mode or update_command
        self.destroy.use_task_objects = not task_mode

        # hide 'destroy' task in the tasks list by removing it from instance
        if not (destroy_command or task_mode):
            del self.destroy

        # set original name for `deploy` method to allow explicit invocation
        if task_mode:
            self.deploy.name = 'deploy'

        # enable following commands only if used custom registry or account
        custom_registry_mode = task_mode or bool(registry or account)
        self.prepare.use_task_objects = custom_registry_mode
        self.push.use_task_objects = custom_registry_mode
        self.upgrade.use_task_objects = custom_registry_mode

    def _set_registry(self, registry):
        self.__dict__['registry'] = docker.Registry(registry)

    def _get_registry(self):
        return self.__dict__.get('registry')

    registry = property(_get_registry, _set_registry)

    def _set_host_registry(self, host_registry):
        self.__dict__['host_registry'] = docker.Registry(host_registry)

    def _get_host_registry(self):
        return self.__dict__.get('host_registry')

    host_registry = property(_get_host_registry, _set_host_registry)

    def _set_ssh_tunnel(self, ssh_tunnel):
        self.__dict__['ssh_tunnel'] = SshTunnel(ssh_tunnel)

    def _get_ssh_tunnel(self):
        return self.__dict__.get('ssh_tunnel')

    ssh_tunnel = property(_get_ssh_tunnel, _set_ssh_tunnel)

    @property
    def image(self):
        return self.service.image

    @fab.task
    @fabricio.skip_unknown_host
    def revert(self):
        """
        revert service container(s) to a previous version
        """
        with self.remote_tunnel():
            self.service.revert()

    @fab.task
    @fabricio.skip_unknown_host
    def migrate(self, tag=None):
        """
        apply new migrations
        """
        with self.remote_tunnel():
            self.service.migrate(
                tag=tag,
                registry=self.host_registry,
                account=self.account,
            )

    @fab.task(name='migrate-back')
    @fabricio.skip_unknown_host
    def migrate_back(self):
        """
        remove previously applied migrations if any
        """
        with self.remote_tunnel():
            self.service.migrate_back()

    @fab.task
    @fabricio.skip_unknown_host
    def backup(self):
        """
        backup service data
        """
        with self.remote_tunnel():
            self.service.backup()

    @fab.task
    @fabricio.skip_unknown_host
    def restore(self, backup_name=None):
        """
        restore service data
        """
        with self.remote_tunnel():
            self.service.restore(backup_name=backup_name)

    @fab.task
    def rollback(self, migrate_back=True):
        """
        rollback service to a previous version (migrate-back -> revert)
        """
        if utils.strtobool(migrate_back):
            self.migrate_back()
        self.revert()

    @fab.hosts()
    @fab.roles()
    @fab.task
    def prepare(self, tag=None):
        """
        download Docker image from the original registry
        """
        if self.registry is None and self.account is None:
            return
        image = self.image[tag]  # type: docker.Image
        if image:
            image.pull(local=True, use_cache=True)

    def delete_dangling_images(self):  # pragma: no cover
        warnings.warn(
            'delete_dangling_images() is deprecated and will be removed in v0.6',  # noqa
            DeprecationWarning,
        )
        warnings.warn(
            'delete_dangling_images() is deprecated and will be removed in v0.6',  # noqa
            RuntimeWarning, stacklevel=self._warnings_stacklevel,
        )
        fabricio.local(dangling_images_delete_command(), ignore_errors=True)

    def push_image(self, tag=None):
        image = self.image[self.registry:tag:self.account]
        fabricio.local(
            'docker push {image}'.format(image=image),
            quiet=False,
            use_cache=True,
        )

    @fab.hosts()
    @fab.roles()
    @fab.task
    def push(self, tag=None):
        """
        push downloaded Docker image to intermediate registry
        """
        if self.registry is None and self.account is None:
            return
        image = self.image[tag]
        if not image:
            return
        proxy_tag = image[self.registry:tag:self.account]
        fabricio.local(
            'docker tag {image} {tag}'.format(
                image=image,
                tag=proxy_tag,
            ),
            use_cache=True,
        )
        self.push_image(tag=tag)
        fabricio.local(
            'docker rmi {tag}'.format(tag=proxy_tag),
            use_cache=True,
        )

    def pull_image(self, tag=None):
        self.service.pull_image(
            tag=tag,
            registry=self.host_registry,
            account=self.account,
        )

    @contextlib.contextmanager
    def remote_tunnel(self):
        with contextlib.ExitStack() as stack:
            if self.ssh_tunnel:
                output = stack.enter_context(contextlib.closing(open(os.devnull, 'w')))  # noqa
                # forward sys.stdout to os.devnull to prevent
                # printing debug messages by fab.remote_tunnel
                stack.enter_context(utils.patch(sys, 'stdout', output))
                stack.enter_context(fab.remote_tunnel(
                    remote_bind_address=self.ssh_tunnel.bind_address,
                    remote_port=self.ssh_tunnel.port,
                    local_host=self.ssh_tunnel.host,
                    local_port=self.ssh_tunnel.host_port,
                ))
            yield

    @fab.task
    @fabricio.skip_unknown_host
    def pull(self, tag=None):
        """
        pull Docker image from the registry
        """
        with self.remote_tunnel():
            self.pull_image(tag=tag)

    @fab.task
    @fabricio.skip_unknown_host
    def update(self, tag=None, force=False):
        """
        update service to a new version
        """
        with self.remote_tunnel():
            updated = self.service.update(
                tag=tag,
                registry=self.host_registry,
                account=self.account,
                force=utils.strtobool(force),
            )
        if updated is False:
            fabricio.log('No changes detected, update skipped.')

    @fab.task
    def upgrade(self, tag=None, force=False, backup=False, migrate=True):
        """
        upgrade service to a new version (backup -> pull -> migrate -> update)
        """
        if utils.strtobool(backup):
            self.backup()
        self.pull(tag=tag)
        if utils.strtobool(migrate):
            self.migrate(tag=tag)
        self.update(tag=tag, force=force)

    @fab.hosts()
    @fab.roles()
    @fab.task(
        default=True,
        # mock another task name to exclude this task from the tasks list
        name='rollback',
    )
    def deploy(self, tag=None, force=False, backup=False, migrate=True):
        """
        full service deploy (prepare -> push -> upgrade)
        """
        self.prepare(tag=tag)
        self.push(tag=tag)
        fab.execute(
            self.upgrade,
            tag=tag,
            force=force,
            backup=backup,
            migrate=migrate,
        )

    class DestroyTask(Task, Tasks):
        """
        delete service, 'destroy.confirm' skips confirmation dialog
        """

        name = 'destroy'

        def __init__(self, tasks, name=None):
            super(DockerTasks.DestroyTask, self).__init__()

            # We need to be sure that `default()` will be at first place
            # every time when vars(self) is being invoked.
            # This is necessary to exclude `default` from the list of task
            # because of it's already there as default task.
            # See `fabric.main.extract_tasks()` for details
            self.__dict__ = utils.PriorityDict(self.__dict__, priority=['default'])

            self.tasks = tasks
            self.name = name or self.name

            # hide actual 'destroy' task from the tasks list
            self.run.use_task_objects = False

        @fab.task(
            default=True,
            # mock another task name to exclude this task from the tasks list
            name='confirm',
        )
        def default(self, *args, **kwargs):
            """
            delete service
            """
            with utils.patch(fab.env, 'parallel', False):
                message = 'Are you sure you want to destroy {name} service?'
                if not console.confirm(
                    message.format(name=colors.red(self.tasks.service.name)),
                    default=False,
                ):
                    fab.abort('Aborted')
            return self.confirm(*args, **kwargs)

        @fab.task
        def confirm(self, *args, **kwargs):
            """
            delete service skipping confirmation dialog
            """
            self.run.hosts = self.hosts
            self.run.roles = self.roles
            return fab.execute(self.run, *args, **kwargs)

        @fab.task(name='destroy')
        def run(self, *args, **kwargs):
            with self.tasks.remote_tunnel():
                return self.tasks.service.destroy(*args, **kwargs)

        def __details__(self):
            with utils.patch(
                six.get_method_function(self.tasks.service.destroy),
                '__doc__',
                self.__doc__ + (self.tasks.service.destroy.__doc__ or ''),
            ):
                return get_task_details(self.tasks.service.destroy)


class ImageBuildDockerTasks(DockerTasks):

    _warnings_stacklevel = 3

    def __init__(self, service=None, build_path='.', **kwargs):
        super(ImageBuildDockerTasks, self).__init__(service, **kwargs)
        self.build_path = build_path
        self.prepare.use_task_objects = True
        self.push.use_task_objects = True
        self.upgrade.use_task_objects = True

    @fab.hosts()
    @fab.roles()
    @fab.task
    def prepare(self, tag=None, **kwargs):
        """
        build Docker image (see 'docker build --help' for available options)
        """
        for key, value in kwargs.items():
            with contextlib.suppress(ValueError):
                kwargs[key] = utils.strtobool(value)
        options = utils.Options(kwargs)
        image = self.image[self.registry:tag:self.account]  # type: docker.Image
        image.build(
            local=True,
            build_path=self.build_path,
            options=options,
            use_cache=True,
        )

    @fab.hosts()
    @fab.roles()
    @fab.task
    def push(self, tag=None):
        """
        push built Docker image to the registry
        """
        self.push_image(tag=tag)



import json
import warnings

import contextlib2 as contextlib

import fabricio

from fabricio import utils

from .base import BaseService, Option, Attribute, ServiceError


class ContainerError(ServiceError):
    pass


class ContainerNotFoundError(ContainerError):
    pass


class Container(BaseService):

    command = Attribute()

    stop_timeout = Attribute(default=10)

    user = Option(safe=True)
    publish = Option()
    env = Option(safe=True)
    label = Option(safe=True)
    volume = Option(safe=True)
    link = Option(safe=True)
    add_host = Option(name='add-host', safe=True)
    network = Option(name='net', safe=True)
    restart = Option()
    stop_signal = Option(name='stop-signal', safe=True)

    @utils.default_property
    def info(self):
        command = 'docker inspect --type container {container}'
        info = fabricio.run(
            command.format(container=self),
            abort_exception=ContainerNotFoundError,
        )
        return json.loads(info)[0]

    def delete(
        self,
        delete_image=False,
        delete_dangling_volumes=None,  # deprecated
        **options
    ):
        options = utils.Options(options)

        if delete_dangling_volumes is not None:  # pragma: no cover
            warnings.warn(
                '`delete_dangling_volumes` parameter is deprecated '
                'and will be removed in v0.6, use `volumes` instead',
                RuntimeWarning, stacklevel=2,
            )
            options.setdefault('volumes', delete_dangling_volumes)

        delete_image_callback = delete_image and self.image.get_delete_callback()

        options.setdefault('volumes', True)  # default option
        fabricio.run('docker rm {options} {container}'.format(
            container=self,
            options=options,
        ))

        if delete_image_callback:
            delete_image_callback()

    def run(self, tag=None, registry=None, account=None):
        self.image[registry:tag:account].run(
            command=self.command,
            temporary=False,
            name=self,
            options=self.options,
        )

    def execute(
        self,
        command=None,
        quiet=True,
        use_cache=False,
        options=(),
    ):
        if not command:
            raise ValueError('Must provide command to execute')

        options = utils.Options(options)
        options.setdefault('tty', True)
        options.setdefault('interactive', True)

        exec_command = 'docker exec {options} {container} {command}'
        return fabricio.run(
            exec_command.format(
                container=self,
                command=command,
                options=options,
            ),
            quiet=quiet,
            use_cache=use_cache,
        )

    def start(self):
        command = 'docker start {container}'
        fabricio.run(command.format(container=self))

    def stop(self, timeout=None):
        if timeout is None:
            timeout = self.stop_timeout
        command = 'docker stop --time {timeout} {container}'
        fabricio.run(command.format(container=self, timeout=timeout))

    def reload(self, timeout=None):
        if timeout is None:
            timeout = self.stop_timeout
        command = 'docker restart --time {timeout} {container}'
        fabricio.run(command.format(container=self, timeout=timeout))

    def rename(self, new_name):
        command = 'docker rename {container} {new_name}'
        fabricio.run(command.format(container=self, new_name=new_name))
        self.name = new_name

    def signal(self, signal):
        command = 'docker kill --signal {signal} {container}'
        fabricio.run(command.format(container=self, signal=signal))

    @property
    def image_id(self):
        return self.info['Image']

    def update(self, tag=None, registry=None, account=None, force=False):
        if not force:
            try:
                if self.image_id == self.image[registry:tag:account].info['Id']:
                    self.start()  # force starting container
                    return False
            except ContainerNotFoundError:
                pass
        obsolete_container = self.get_backup_version()
        with contextlib.suppress(fabricio.Error):
            obsolete_container.delete(delete_image=True)
        try:
            backup_container = self.fork()
            backup_container.rename(obsolete_container.name)
        except fabricio.Error:
            pass  # current container not found
        else:
            backup_container.stop()
        self.run(tag=tag, registry=registry, account=account)
        return True

    def revert(self):
        backup_container = self.get_backup_version()
        try:
            backup_container.info
        except ContainerNotFoundError:
            raise ContainerError('backup container not found')
        self.stop()
        backup_container.start()
        self.delete(delete_image=True)
        backup_container.rename(self.name)

    def get_backup_version(self):
        return self.fork(name='{container}_backup'.format(container=self))

    def destroy(self):
        self.delete(force=True, delete_image=True)
        with contextlib.suppress(fabricio.Error):
            self.get_backup_version().delete(force=True, delete_image=True)



from fabricio.operations import local, log, move_file, remove_file, run
from fabricio.operations import Error, host_errors
from fabricio.decorators import skip_unknown_host, once_per_task
from fabricio.tasks import infrastructure

VERSION = (0, 5, 4)

__version__ = '.'.join(map(str, VERSION))



import collections
import itertools

import six

from six.moves import map, filter

from fabricio import docker, kubernetes, utils, once_per_task


class Migration(six.text_type):

    def __init__(self, *args, **kwargs):
        super(Migration, self).__init__(**kwargs)
        self.app, _, self.name = self.partition('.')


class DjangoMixin(docker.BaseService):
    """
    Be sure you use proper Dockerfile's WORKDIR directive
    (or another alternative) which points to the directory where
    manage.py placed
    """

    @staticmethod
    def _migrate(image, options):
        image.run(
            'python manage.py migrate --noinput',
            quiet=False,
            options=options,
        )

    @once_per_task
    def migrate(self, tag=None, registry=None, account=None):
        self._migrate(
            image=self.image[registry:tag:account],
            options=self.safe_options,
        )

    @staticmethod
    def _get_parent_migration(migration, migrations):
        migrations = iter(migrations)
        any(map(migration.__eq__, migrations))  # skip later migrations
        for parent_migration in migrations:
            if migration.app == parent_migration.app:
                return parent_migration
        return Migration(migration.app + '.zero')

    def get_revert_migrations(self, current_migrations, backup_migrations):
        current_migrations, all_migrations = itertools.tee(reversed(list(map(
            Migration,
            filter(None, current_migrations.splitlines()),
        ))))
        all_migrations = utils.OrderedSet(all_migrations)

        backup_migrations = reversed(list(map(
            Migration,
            filter(None, backup_migrations.splitlines()),
        )))

        revert_migrations = collections.OrderedDict()

        while True:
            while True:
                backup_migration = next(backup_migrations, None)
                if not backup_migration or backup_migration in all_migrations:
                    break
            for current_migration in current_migrations:
                if current_migration == backup_migration:
                    break
                revert_migration = self._get_parent_migration(
                    current_migration,
                    migrations=all_migrations,
                )
                revert_migrations[current_migration.app] = revert_migration

            if backup_migration is None:
                return revert_migrations.values()

    @once_per_task
    def migrate_back(self):
        migrations_command = (
            'python manage.py showmigrations --plan '  # execute plan
            '| egrep \'^\\[X\\]\' '  # show only applied migrations
            '| awk \'{print $2}\' '  # take migration name
            '&& test ${PIPESTATUS[0]} -eq 0'  # fail if couldn't execute plan
        )
        image = self.image
        options = self.safe_options

        with utils.patch(self, 'info', self.info, force_delete=True):
            current_migrations = image.run(
                migrations_command,
                options=options,
            )
            backup = self.get_backup_version()
            with utils.patch(backup, 'info', backup.info, force_delete=True):
                backup_migrations = backup.image.run(
                    migrations_command,
                    options=options,
                )

                for migration in self.get_revert_migrations(
                    current_migrations,
                    backup_migrations,
                ):
                    command = (
                        'python manage.py migrate --no-input {app} {migration}'
                    ).format(
                        app=migration.app,
                        migration=migration.name,
                    )
                    image.run(command, quiet=False, options=options)

                self._migrate(backup.image, options)


class DjangoContainer(docker.Container, DjangoMixin):
    pass


class DjangoService(docker.Service, DjangoMixin):
    pass


class DjangoStack(docker.Stack, DjangoMixin):
    pass


class DjangoKubernetes(kubernetes.Configuration, DjangoMixin):
    pass



import json
import warnings

import docker.auth
import docker.utils
import six

from functools import partial

import fabricio

from fabricio import utils


class ImageError(fabricio.Error):
    pass


class ImageNotFoundError(ImageError):
    pass


class Registry(six.text_type):

    def __new__(cls, value=None, *args, **kwargs):
        if value is None:
            return None
        return super(Registry, cls).__new__(cls, value, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        super(Registry, self).__init__(**kwargs)
        self.host, _, port = self.partition(':')
        self.port = port and int(port)


class Image(object):

    name = None

    tag = None

    registry = None

    use_digest = False

    @property
    def temp_tag(self):
        return 'fabricio-temp-image:' + self.name.rsplit('/')[-1]

    def __new__(cls, name=None, tag=None, registry=None):
        if isinstance(name, Image):
            return name[registry:tag]
        return super(Image, cls).__new__(cls)

    def __init__(self, name=None, tag=None, registry=None):
        if name is not None and not isinstance(name, Image):
            _registry, _name, _tag = self.parse_image_name(name)
            self.name = _name
            self.tag = tag or _tag or 'latest'  # TODO 'latest' is unnecessary
            self.registry = Registry(registry or _registry)
            self.use_digest = not tag and '@' in name
        self.field_names = {}  # descriptor's cache
        self.service = None

    def __str__(self):
        if self.service is not None:
            image_id = getattr(self.service, 'image_id', None)
            if image_id:
                return image_id
        return super(Image, self).__str__()

    def __repr__(self):
        if not self.name:
            raise ImageError('image name is not set or empty')
        tag_separator = '@' if self.use_digest else ':'
        registry = self.registry and '{0}/'.format(self.registry) or ''
        tag = self.tag and '{0}{1}'.format(tag_separator, self.tag) or ''
        return '{registry}{name}{tag}'.format(
            registry=registry,
            name=self.name,
            tag=tag,
        )

    def __bool__(self):
        try:
            return bool(repr(self))
        except ImageError:
            return False

    def __nonzero__(self):
        return self.__bool__()

    def __get__(self, service, owner_cls):
        if service is None:
            return self
        field_name = self.get_field_name(owner_cls)
        if field_name not in service.__dict__:
            image = service.__dict__[field_name] = self[:]
            image.service = service
        return service.__dict__[field_name]

    def __set__(self, service, image):
        field_name = self.get_field_name(type(service))
        image = self.__class__(image)
        image.service = service
        service.__dict__[field_name] = image

    def __getitem__(self, item):
        if isinstance(item, slice):
            registry, tag, account = item.start, item.stop, item.step
        else:
            registry, tag, account = None, item, None

        use_digest = self.use_digest

        # tag can override image registry, name and/or digest
        _registry, _name, _tag = self.parse_image_name(tag)
        if not _tag:
            if _registry:
                _tag = 'latest'
            else:
                _tag, _name = _name, None
        if _tag:
            use_digest = _name and tag and '@' in tag

        registry = _registry or registry or self.registry
        name = _name or account and self.name and '{account}/{name}'.format(
            account=account,
            name=self.name.split('/')[-1],
        ) or self.name
        tag = _tag or tag or self.tag

        if use_digest:
            name = '{name}@{digest}'.format(name=name, digest=tag)
            tag = None

        return self.__class__(name=name, tag=tag, registry=registry)

    def get_field_name(self, owner_cls):
        field_name = self.field_names.get(owner_cls)
        if field_name is None:
            for attr in dir(owner_cls):
                if getattr(owner_cls, attr) is self:
                    if field_name is not None:
                        raise ValueError(
                            'Same instance of Image used for more than one '
                            'attribute of class {cls}'.format(
                                cls=owner_cls.__name__,
                            )
                        )
                    self.field_names[owner_cls] = field_name = attr
        return field_name

    @staticmethod
    def parse_image_name(image):
        if not image:
            return None, None, None
        repository, tag = docker.utils.parse_repository_tag(image)
        registry, name = docker.auth.resolve_repository_name(repository)
        if registry == docker.auth.INDEX_NAME:
            registry = None
        return registry, name, tag

    @property
    def digest(self):
        if not self.use_digest:
            for repo_digest in self.info.get('RepoDigests', ()):
                return repo_digest
            raise ImageError('image has no digest')
        return repr(self)

    @utils.default_property
    def info(self):
        command = 'docker inspect --type image {image}'
        info = fabricio.run(
            command.format(image=self),
            abort_exception=ImageNotFoundError,
        )
        return json.loads(info)[0]

    def get_delete_callback(self, force=False):
        command = 'docker rmi {force}{image}'
        force = force and '--force ' or ''
        return partial(
            fabricio.run,
            command.format(image=self, force=force),
            ignore_errors=True,
        )

    def delete(self, force=False, ignore_errors=True):
        delete_callback = self.get_delete_callback(force=force)
        return delete_callback(ignore_errors=ignore_errors)

    @classmethod
    def make_container_options(
        cls,
        temporary=None,
        name=None,
        options=(),
    ):
        return utils.Options(
            options,
            name=name,
            rm=temporary,
            tty=temporary,
            interactive=temporary,
            detach=temporary is not None and not temporary,
        )

    def run(
        self,
        command=None,
        name=None,
        temporary=True,
        options=(),
        quiet=True,
    ):
        run_command = 'docker run {options} {image} {command}'
        return fabricio.run(
            run_command.format(
                image=self,
                command=command or '',
                options=self.make_container_options(
                    temporary=temporary,
                    name=name,
                    options=options,
                ),
            ),
            quiet=quiet,
        )

    def create(self, command=None, name=None, options=()):  # pragma: no cover
        warnings.warn('Image.create() is deprecated', DeprecationWarning)
        warnings.warn(
            'Image.create() is deprecated',
            RuntimeWarning, stacklevel=2,
        )
        run_command = 'docker create {options} {image} {command}'.rstrip()
        return fabricio.run(
            run_command.format(
                image=self,
                command=command or '',
                options=self.make_container_options(name=name, options=options),
            ),
        )

    def pull(self, local=False, use_cache=False, ignore_errors=False):
        run = fabricio.local if local else fabricio.run
        run = partial(run, use_cache=use_cache, ignore_errors=ignore_errors)
        run_ignore_errors = partial(run, ignore_errors=True)

        image = six.text_type(self)

        run_ignore_errors(
            'docker tag {image} {tag} '
            '&& docker rmi {image}'.format(image=image, tag=self.temp_tag)
        )
        pull_result = run('docker pull ' + image, quiet=False)
        if pull_result.succeeded:
            run_ignore_errors('docker rmi {tag}'.format(tag=self.temp_tag))

    def build(self, local=False, build_path='.', options=None, use_cache=False):
        if local:
            run = fabricio.local
            run_capture_output = partial(run, capture=True)
        else:
            run = run_capture_output = fabricio.run
        run = partial(run, use_cache=use_cache)
        run_capture_output = partial(run_capture_output, use_cache=use_cache)
        run_ignore_errors = partial(run, ignore_errors=True)

        image = six.text_type(self)
        options = options or utils.Options()
        options['tag'] = image

        # default options
        options.setdefault('pull', 1)
        options.setdefault('force-rm', 1)

        with utils.patch(fabricio, 'run', run_capture_output):
            try:
                old_parent_id = self.info['Parent']
            except ImageNotFoundError:
                old_parent_id = ''

        run_ignore_errors(
            'docker tag {image} {tag} '
            '&& docker rmi {image}'.format(image=image, tag=self.temp_tag)
        )
        run(
            'docker build {options} {build_path}'.format(
                options=options,
                build_path=build_path,
            ),
            quiet=False,
        )
        run_ignore_errors('docker rmi {tag} {old_parent}'.format(
            tag=self.temp_tag,
            old_parent=old_parent_id,
        ))
